# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

import click

from clapper.click import ResourceOption, verbosity_option
from clapper.logging import setup

from .click import ConfigCommand

# avoids X11/graphical desktop requirement when creating plots
__import__("matplotlib").use("agg")

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


@click.command(
    entry_point_group="ptbench.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Runs evaluation on an existing prediction output:

   .. code:: sh

      ptbench evaluate -vv --predictions=path/to/predictions.json --output-folder=path/to/results

2. Runs evaluation on an existing prediction output, tune threshold a priori on the `validation` set:

   .. code:: sh

      ptbench evaluate -vv --predictions=path/to/predictions.json --output-folder=path/to/results --threshold=validation
""",
)
@click.option(
    "--predictions",
    "-p",
    help="Path where predictions are currently stored",
    required=True,
    type=click.Path(
        file_okay=True, dir_okay=False, writable=True, path_type=pathlib.Path
    ),
    cls=ResourceOption,
)
@click.option(
    "--output-folder",
    "-o",
    help="Path where to store the analysis result (created if does not exist)",
    required=True,
    default="results",
    type=click.Path(file_okay=False, dir_okay=True, path_type=pathlib.Path),
    cls=ResourceOption,
)
@click.option(
    "--threshold",
    "-t",
    help="""This value is used to define positives and negatives from
    probability outputs in predictions, and report performance measures on
    **binary** classification tasks. It should either come from the training
    set or a separate validation set to avoid biasing the analysis.
    Optionally, if you provide a multi-split set of predictions as input, this
    may also be the name of an existing split (e.g. ``validation``) from which
    the threshold will be estimated (by calculating the threshold leading to
    the highest F1-score on that set) and then applied to the subsequent
    sets.  This value is not used for multi-class classification tasks.""",
    default=0.5,
    show_default=False,
    required=True,
    type=click.STRING,
    cls=ResourceOption,
)
@verbosity_option(logger=logger, cls=ResourceOption, expose_value=False)
def evaluate(
    predictions: pathlib.Path,
    output_folder: pathlib.Path,
    threshold: str | float,
    **_,  # ignored
) -> None:
    """Evaluates predictions (from a model) on a classification task."""

    import json
    import typing

    import matplotlib.figure

    from matplotlib.backends.backend_pdf import PdfPages

    from ..engine.evaluator import (
        aggregate_pr,
        aggregate_roc,
        aggregate_summaries,
        run_binary,
    )

    with predictions.open("r") as f:
        predict_data = json.load(f)

    if threshold in predict_data:
        # it is the name of a split
        # first run evaluation for reference dataset
        from ..engine.evaluator import maxf1_threshold

        use_threshold = maxf1_threshold(predict_data[threshold])
        logger.info(f"Setting --threshold={use_threshold:.5f}")

    else:
        # we try to convert it to float and complain if that is not possible
        try:
            use_threshold = float(threshold)
        except ValueError:
            raise click.BadParameter(
                f"""The value of --threshold=`{threshold}` does not match one
                of the database split names ({', '.join(predict_data.keys())})
                or can be converted to float. Check your input."""
            )

    results: dict[
        str,
        tuple[
            dict[str, typing.Any],
            dict[str, matplotlib.figure.Figure],
            dict[str, typing.Any],
        ],
    ] = dict()
    for k, v in predict_data.items():
        logger.info(f"Analyzing split `{k}`...")
        results[k] = run_binary(
            name=k,
            predictions=v,
            threshold_a_priori=use_threshold,
        )

    rows = [v[0] for v in results.values()]
    table = aggregate_summaries(rows, fmt="rst")
    click.echo(table)

    if output_folder is not None:
        output_folder.mkdir(parents=True, exist_ok=True)

        table_path = output_folder / "summary.rst"

        logger.info(f"Saving measures at `{table_path}`...")
        with table_path.open("w") as f:
            f.write(table)

        figure_path = output_folder / "plots.pdf"
        logger.info(f"Saving figures at `{figure_path}`...")

        with PdfPages(figure_path) as pdf:
            pr_curves = {
                k: v[2]["precision_recall"] for k, v in results.items()
            }
            pr_fig = aggregate_pr(pr_curves)
            pdf.savefig(pr_fig)

            roc_curves = {k: v[2]["roc"] for k, v in results.items()}
            roc_fig = aggregate_roc(roc_curves)
            pdf.savefig(roc_fig)

            # order ready-to-save figures by type instead of split
            figures = {k: v[1] for k, v in results.items()}
            keys = next(iter(figures.values())).keys()
            figures_by_type = {
                k: [v[k] for v in figures.values()] for k in keys
            }

            for group_figures in figures_by_type.values():
                for f in group_figures:
                    pdf.savefig(f)
